from datetime import datetime
from typing import Any

from hatchet_sdk.runnables.workflow import Workflow
from hatchet_sdk.runnables.types import WorkflowConfig
from hatchet_sdk.clients.listeners.run_event_listener import StepRunEventType
from hatchet_sdk.clients.rest.models.v1_workflow_run_details import V1WorkflowRunDetails
from hatchet_sdk.clients.rest.models.v1_task_summary import V1TaskSummary
from hatchet_sdk.clients.rest.models.v1_task_status import V1TaskStatus
from pydantic import (
    BaseModel,
    HttpUrl,
    ConfigDict,
    field_serializer,
)
from pydantic.json_schema import SkipJsonSchema
from fastapi import FastAPI
from typer import Typer

from knoc.utils import get_readable_error


class Plugin(BaseModel):
    id: str
    module: str
    description: str


class APIPlugin(Plugin):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    api: SkipJsonSchema[FastAPI]

    @field_serializer("api")
    def serialize_api(self, api: FastAPI, _info):
        pass


class CLIPlugin(Plugin):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    cli: SkipJsonSchema[Typer]

    @field_serializer("cli")
    def serialize_cli(self, cli: Typer, _info):
        pass


def serialize_workflow_config(config: WorkflowConfig) -> dict[str, Any]:
    return config.model_json_schema()


class WorkflowMixin(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    wf: SkipJsonSchema[Workflow]

    @field_serializer("wf")
    def serialize_workflow(self, wf: Workflow, _info):
        return {
            "config": wf.config.model_dump(
                mode="json",
                exclude_defaults=True,
                fallback=serialize_workflow_config,
            ),
            "id": wf.id,
            "name": wf.name,
            "service_name": wf.service_name,
            "tasks": [task.name for task in wf.tasks],
        }


class WorkflowPlugin(WorkflowMixin, Plugin):
    pass


class WorkflowTriggerResult(BaseModel):
    wfRunId: str


class WorkflowRunResult(WorkflowMixin):
    wfRunId: str
    result: dict[str, Any] = {}
    error: dict[str, Any] = {}
    payload: str | None = None
    input: dict[str, Any] | None = None
    finishedAt: datetime | None
    status: V1TaskStatus
    duration: int | None

    @classmethod
    def from_run_details(
        cls, wf: Workflow, run_details: V1WorkflowRunDetails
    ) -> "WorkflowRunResult":
        failure_task_name = wf._on_failure_task and wf._on_failure_task.name
        return cls(
            wf=wf,
            wfRunId=run_details.run.metadata.id,
            result={
                child.action_id.split(":")[-1]: child.output
                for child in run_details.tasks
                if child.action_id
                and child.status == V1TaskStatus.COMPLETED
                and child.action_id.split(":")[1] != failure_task_name
            },
            error={
                child.action_id.split(":")[-1]: get_readable_error(child.error_message)
                for child in run_details.tasks or []
                if child.action_id
                and child.status == V1TaskStatus.FAILED
                and child.error_message
            },
            input=run_details.run.input,
            status=run_details.run.status,
            finishedAt=run_details.run.finished_at,
            duration=run_details.run.duration,
        )

    @classmethod
    def from_task_summary(cls, wf: Workflow, run: V1TaskSummary) -> "WorkflowRunResult":
        failure_task_name = wf._on_failure_task and wf._on_failure_task.name
        return cls(
            wf=wf,
            wfRunId=run.workflow_run_external_id,
            result={
                child.action_id.split(":")[-1]: child.output
                for child in run.children or []
                if child.action_id
                and child.status == V1TaskStatus.COMPLETED
                and child.action_id.split(":")[1] != failure_task_name
            },
            error={
                child.action_id.split(":")[-1]: get_readable_error(child.error_message)
                for child in run.children or []
                if child.action_id
                and child.status == V1TaskStatus.FAILED
                and child.error_message
            },
            input=run.input,
            status=run.status,
            finishedAt=run.finished_at,
            duration=run.duration,
        )


class SettingsPlugin(Plugin): ...


class Plugins(BaseModel):
    api: dict[str, APIPlugin] = {}
    cli: dict[str, CLIPlugin] = {}
    workflow: dict[str, WorkflowPlugin] = {}
    settings: dict[str, SettingsPlugin] = {}


class ServerInfo(BaseModel):
    version: str
    hatchet_url: HttpUrl
    api_doc: str
    plugins: Plugins = Plugins()


class HatchetStepEvent(BaseModel):
    wfId: str
    wfRunId: str
    type: StepRunEventType
    payload: str | None = None
    result: dict[str, Any]
