import asyncio
from typing import Annotated
from contextlib import asynccontextmanager
from importlib.metadata import entry_points
import logging

from pydantic import ValidationError
from fastapi import FastAPI, Body, HTTPException, status, WebSocket, WebSocketDisconnect
from fastapi.middleware.cors import CORSMiddleware
from hatchet_sdk.clients.rest.models.worker import Worker
from hatchet_sdk.workflow_run import WorkflowRunRef
from jwt import decode as decode_jwt

from knoc import __version__
from knoc.settings import settings
from knoc.models import (
    ServerInfo,
    APIPlugin,
    WorkflowPlugin,
    WorkflowRunResult,
    WorkflowTriggerResult,
)
from knoc.workflows import workflows
from knoc.hatchet_client import hatchet
from knoc.websocket import manager

# from knoc.workflows.utils import ImportReport
# from knoc.db import get_session

logger = logging.getLogger("knoc.server")


@asynccontextmanager
async def lifespan(app: FastAPI):
    # Load the FastAPI app
    assert app
    logger.debug("App is Starting Up!")
    yield
    logger.info(
        "App is Shutting Down Gracefully!",
        extra={"markup": True},
    )


app = FastAPI(lifespan=lifespan)
app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.server.cors,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

server_info = ServerInfo(
    version=__version__,
    hatchet_url=decode_jwt(
        settings.hatchet_client.token, options={"verify_signature": False}
    )["server_url"],
    api_doc="/docs",
)


def get_plugins():
    for entry_point in entry_points().select(group="knoc.api"):
        try:
            api = entry_point.load()
        except Exception as err:
            logger.warning(
                f"Cannot load settings module {entry_point.name} - see below"
            )
            logger.exception(err)
            continue
        if not isinstance(api, FastAPI):
            logger.warning(f"Settings module {entry_point.name} is not a FastAPI")
            continue
        app.mount("/" + entry_point.name, api)
        server_info.plugins.api[entry_point.value] = APIPlugin(
            id=entry_point.name,
            module=entry_point.module,
            description=api.description,
            api=api,
        )
        logger.info(f"Mounted {entry_point.name}")


server_info.plugins.workflow = workflows

get_plugins()


@app.get("/")
async def root() -> ServerInfo:
    return server_info


@app.get("/wf")
async def list_workflows() -> dict[str, WorkflowPlugin]:
    """List the workflows"""
    # return {k: wf.wf.config for k, wf in workflows.items()}  # type: ignore
    return workflows


@app.post("/wf/{workflow_id}/trigger")
async def trigger_workflow(
    workflow_id: str, input: Annotated[dict, Body()]
) -> WorkflowTriggerResult:
    """Trigger a workflow and return the workflow id"""
    wf = workflows[workflow_id].wf
    try:
        input_ = wf.config.input_validator(**input)
    except ValidationError as err:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=err.title)
    else:
        # Trigger the workflow and track its execution
        wf_run_ref: WorkflowRunRef = await wf.aio_run_no_wait(input=input_)
        asyncio.create_task(manager.track_hatchet_run(wf=wf, wf_run_ref=wf_run_ref))
    return WorkflowTriggerResult(wfRunId=wf_run_ref.workflow_run_id)


@app.post("/wf/{workflow_id}/run")
async def run_workflow(
    workflow_id: str, input: Annotated[dict, Body()]
) -> WorkflowRunResult:
    """Execute a workflow and wait for the result"""
    wf = workflows[workflow_id].wf
    try:
        input_ = wf.config.input_validator(**input)
    except ValidationError as err:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=err.title)
    else:
        wf_run_ref: WorkflowRunRef = await wf.aio_run_no_wait(input=input_)
    return WorkflowRunResult.from_run_details(
        wf, await hatchet.runs.aio_get(wf_run_ref.workflow_run_id)
    )


@app.get("/wf/{workflow_id}/runs")
async def workflow_results(workflow_id: str) -> list[WorkflowRunResult]:
    """Get the results of the runs for the workflow"""
    ## Confusion between own workflow plugins,
    ## hatchet_sdk.runnables.workflow.Workflow and hatchet_sdk.clients.rest.models.workflow.Workflow
    # wf = await hatchet.workflows.aio_get(workflow_id) -> get hatchet_sdk.clients.rest.models.workflow.Workflow
    # Easy to go with own workflow plugin:
    wf_plugin = [ww for ww in workflows.values() if ww.wf.id == workflow_id][0]
    runs = await hatchet.runs.aio_list(workflow_ids=[workflow_id])
    return [
        WorkflowRunResult.from_task_summary(wf_plugin.wf, run)
        for run in runs.rows
        if run.children
    ]


@app.get("/worker")
async def list_workers() -> list[Worker] | None:
    """List the workers"""
    workers = await hatchet.workers.aio_list()
    return workers.rows


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await manager.connect(websocket)
    try:
        while True:
            request = await websocket.receive_json()
            await manager.process_request(request, websocket)
    except WebSocketDisconnect:
        manager.disconnect(websocket)
