from typing import Literal
import logging

from pydantic import BaseModel
from fastapi import WebSocket
from hatchet_sdk.workflow_run import WorkflowRunRef
from hatchet_sdk.runnables.workflow import Workflow
from hatchet_sdk.clients.listeners.run_event_listener import StepRunEventType

from knoc.models import WorkflowRunResult, HatchetStepEvent
from knoc.hatchet_client import hatchet

logger = logging.getLogger("knoc.server.ws")


class WSRequest(BaseModel):
    type: Literal["subscribe", "unsubscribe"]
    wfId: str | None = None


class ConnectionManager:
    def __init__(self):
        self.active_connections: list[WebSocket] = []

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket):
        self.active_connections.remove(websocket)

    async def process_request(self, request: WSRequest, websocket: WebSocket):
        # await websocket.send_text(message)
        if request.type == "subscribe":
            websocket.items()

    async def broadcast(self, message: str):
        for connection in self.active_connections:
            await connection.send_text(message)

    async def broadcast_hatchet_event(self, event: HatchetStepEvent):
        event_dump = event.model_dump()
        for connection in self.active_connections:
            await connection.send_json(event_dump)

    async def broadcast_workflow_run_result(self, event: WorkflowRunResult):
        event_dump = event.model_dump(mode="json")
        for connection in self.active_connections:
            await connection.send_json(event_dump)

    async def track_hatchet_run(self, wf: Workflow, wf_run_ref: WorkflowRunRef):
        async for event in wf_run_ref.stream():
            if (
                event.type == StepRunEventType.STEP_RUN_EVENT_TYPE_COMPLETED
                or event.type == StepRunEventType.STEP_RUN_EVENT_TYPE_FAILED
            ):
                await self.broadcast_workflow_run_result(
                    WorkflowRunResult.from_run_details(
                        wf, await hatchet.runs.aio_get(wf_run_ref.workflow_run_id)
                    )
                )
            else:
                # TODO: Send a notification
                logger.warning(f"Event type {event.type} for wf {wf_run_ref}")


manager = ConnectionManager()
