import logging
from importlib.metadata import entry_points
from datetime import datetime

from pydantic import BaseModel

from hatchet_sdk.runnables.workflow import Workflow
from hatchet_sdk.runnables.types import StepType
from hatchet_sdk.clients.rest.models.v1_task_status import V1TaskStatus
from hatchet_sdk import Context

from knoc.models import WorkflowPlugin
from knoc.websocket import manager
from knoc.models import WorkflowRunResult
from knoc.utils import get_readable_errors

logger = logging.getLogger("knoc.workflows")


class WorkflowError(Exception):
    pass


workflows: dict[str, WorkflowPlugin] = {}


async def on_failure(input: BaseModel, ctx: Context):
    assert ctx.workflow_id
    await manager.broadcast_workflow_run_result(
        event=WorkflowRunResult(
            wf=get_workflow_by_hatchet_id(ctx.workflow_id),
            wfRunId=ctx.workflow_run_id,
            result={},
            error=get_readable_errors(ctx.task_run_errors),
            finishedAt=datetime.now(),
            status=V1TaskStatus.FAILED,
            duration=None,
        )
    )


def get_workflow_by_hatchet_id(workflow_id: str) -> Workflow:
    return [
        wf_plugin.wf
        for wf_plugin in workflows.values()
        if wf_plugin.wf.id == workflow_id
    ][0]


def get_plugins():
    for entry_point in entry_points().select(group="knoc.workflow"):
        try:
            wf = entry_point.load()
        except Exception as err:
            logger.warning(f"Cannot load CLI module {entry_point.name} - see below")
            logger.exception(err)
            continue
        if not isinstance(wf, Workflow):
            logger.warning(f"CLI module {entry_point.name} is not a Typer")
            continue

        ## Add a task for catching failures
        if (
            len([task.type for task in wf.tasks if task.type == StepType.ON_FAILURE])
            == 0
        ):
            wf.on_failure_task()(on_failure)
        workflows[entry_point.value] = WorkflowPlugin(
            id=entry_point.name,
            module=entry_point.module,
            description=wf.config.description or wf.name.capitalize().replace("-", " "),
            wf=wf,
        )


get_plugins()
